"""
Load models for ASR
"""

import os
from pathlib import Path

import tensorflow as tf
from at16k.utils.text_encoder import SubwordTextEncoder

class Model():
    """ASR Model
    Two propoerties/methods are exposed:
    1) pred_fn: converts a waveform to a list of ids
    2) output_fn: converts list of ids to text
    """
    def __init__(self, name):
        self.name = name
        self._pred_fn = self._build_pred_fn()
        self._output_fn = self._build_output_fn()

    @property
    def pred_fn(self):
        """
        Converts waveform to list of ids
        """
        return self._pred_fn

    @property
    def output_fn(self):
        """
        Converts list of ids to text
        """
        return self._output_fn

    @staticmethod
    def _get_model_dir(name):
        sys_home_dir = str(Path.home())
        model_dir = os.path.join(sys_home_dir, '.at16k', name)
        assert os.path.exists(model_dir), ('%s model does not exist at %s' % (name, model_dir))
        return model_dir

    def _build_pred_fn(self):
        name = self.name
        model_dir = self._get_model_dir(name)
        model_path = os.path.join(model_dir, 'export')
        sub_dirs = os.listdir(model_path)
        sub_dirs.sort()
        latest_model_path = os.path.join(model_path, sub_dirs[-1])
        pred_fn = tf.contrib.predictor.from_saved_model(latest_model_path)
        return pred_fn

    def _build_output_fn(self):
        name = self.name
        model_dir = self._get_model_dir(name)
        model_path = os.path.join(model_dir, 'vocab', 'bpe.1000.t2t')
        vocab_model = SubwordTextEncoder(filename=model_path)
        return vocab_model.decode



"""
Audio segment
"""


class Segment():
    """
    Audio segment
    """

    def __init__(self, waveform, boundaries, sample_rate, channel):
        self._waveform = waveform
        self._boundaries = boundaries
        self._sample_rate = sample_rate
        self._channel = channel

    @property
    def waveform(self):
        """
        Segment waveform
        """
        return self._waveform

    @property
    def boundaries(self):
        """
        Segment begin and end timestamp (in seconds)
        """
        return self._boundaries

    @property
    def sample_rate(self):
        """
        Sample rate
        """
        return self._sample_rate

    @property
    def channel(self):
        """
        Channel number (starts from 0)
        """
        return self._channel



import os
import shutil
import urllib.request
from pathlib import Path
import sys
import progressbar

BASE_URL = 'https://storage.googleapis.com/at16k-ce/models'
AVAILABLE_MODELS = ['en_8k', 'en_16k']

PROGRESS_BAR = None


def show_progress(block_num, block_size, total_size):
    global PROGRESS_BAR
    if PROGRESS_BAR is None:
        PROGRESS_BAR = progressbar.ProgressBar(maxval=total_size)
        PROGRESS_BAR.start()

    downloaded = block_num * block_size
    if downloaded < total_size:
        PROGRESS_BAR.update(downloaded)
    else:
        PROGRESS_BAR.finish()
        PROGRESS_BAR = None


def setup_home():
    """
    Init home directory to store assets and models
    """
    home_dir = str(Path.home())
    at16k_model_dir = os.path.join(home_dir, '.at16k')
    if not os.path.exists(at16k_model_dir):
        os.mkdir(at16k_model_dir)
    return at16k_model_dir


def download_model(remote_path, local_path):
    """
    Download file
    """
    if not os.path.exists(local_path):
        print('Downloading from %s' % remote_path)
        urllib.request.urlretrieve(remote_path, local_path, show_progress)


def unarchive(local_path, base_dir):
    """
    Unarchive zipped file
    """
    shutil.unpack_archive(local_path, base_dir)


def main():
    assert len(
        sys.argv) > 1, ('Please specify model name: one of en_8k, en_16k, all')
    name = sys.argv[1]
    if name == 'all':
        name = AVAILABLE_MODELS
    else:
        assert name in AVAILABLE_MODELS, (
            'Please specify a valid model name: one of en_8k, en_16k, all')
        name = [name]
    base_dir = setup_home()
    for item in name:
        file_name = '%s.tar.gz' % item
        remote_path = os.path.join(BASE_URL, file_name)
        local_path = os.path.join(base_dir, file_name)
        download_model(remote_path, local_path)
        unarchive(local_path, base_dir)
        os.remove(local_path)
        print('Downloaded model: %s' % item)


if __name__ == '__main__':
    main()



"""
Audio file handler
"""

import scipy.io.wavfile as wavfile
import numpy as np

class Media():
    """
    Media: I/O functionality to read/write audio files
    """

    def __init__(self, file_path):
        self.file_path = file_path

    @property
    def sample_rate(self):
        """
        Sample rate of audio file
        """
        sample_rate, _ = wavfile.read(self.file_path)
        return sample_rate

    @property
    def waveform(self):
        """
        Raw waveform of entire audio file
        """
        _, waveform = wavfile.read(self.file_path)
        shape = np.shape(waveform)
        if len(shape) > 1:
            waveform = np.mean(waveform, axis=1)
        return waveform

    @property
    def duration(self):
        """
        Length of the audio file (in seconds)
        """
        return len(self.waveform)/float(self.sample_rate)



# coding=utf-8
# Copyright 2019 The Tensor2Tensor Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A simple invertible tokenizer.

Converts from a unicode string to a list of tokens
(represented as Unicode strings).

This tokenizer has the following desirable properties:
 - It is invertible.
 - Alphanumeric characters are broken away from non-alphanumeric characters.
 - A single space between words does not produce an extra token.
 - The full Unicode punctuation and separator set is recognized.

The tokenization algorithm is as follows:

1.    Split the text into a list of tokens, splitting at every boundary of an
        alphanumeric character and a non-alphanumeric character.    This produces
        a list which alternates between "alphanumeric tokens"
        (strings of alphanumeric characters) and "non-alphanumeric tokens"
        (strings of non-alphanumeric characters).

2.    Remove every token consisting of a single space, unless it is
        the very first or very last token in the list.    These tokens are now
        implied by the fact that there are two adjacent alphanumeric tokens.

e.g.    u"Dude - that's so cool."
                -> [u"Dude", u" - ", u"that", u"'", u"s", u"so", u"cool", u"."]
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import sys
import unicodedata
import six
from six.moves import range    # pylint: disable=redefined-builtin
import tensorflow as tf

# Conversion between Unicode and UTF-8, if required (on Python2)
_native_to_unicode = (lambda s: s.decode(
    "utf-8")) if six.PY2 else (lambda s: s)


# This set contains all letter and number characters.
_ALPHANUMERIC_CHAR_SET = set(
    six.unichr(i) for i in range(sys.maxunicode)
    if (unicodedata.category(six.unichr(i)).startswith("L") or
        unicodedata.category(six.unichr(i)).startswith("N")))


def encode(text):
    """Encode a unicode string as a list of tokens.

    Args:
        text: a unicode string
    Returns:
        a list of tokens as Unicode strings
    """
    if not text:
        return []
    ret = []
    token_start = 0
    # Classify each character in the input string
    is_alnum = [c in _ALPHANUMERIC_CHAR_SET for c in text]
    for pos in range(1, len(text)):
        if is_alnum[pos] != is_alnum[pos - 1]:
            token = text[token_start:pos]
            if token != u" " or token_start == 0:
                ret.append(token)
            token_start = pos
    final_token = text[token_start:]
    ret.append(final_token)
    return ret


def decode(tokens):
    """Decode a list of tokens to a unicode string.

    Args:
        tokens: a list of Unicode strings
    Returns:
        a unicode string
    """
    token_is_alnum = [t[0] in _ALPHANUMERIC_CHAR_SET for t in tokens]
    ret = []
    for i, token in enumerate(tokens):
        if i > 0 and token_is_alnum[i - 1] and token_is_alnum[i]:
            ret.append(u" ")
        ret.append(token)
    return "".join(ret)


def _read_filepattern(filepattern, max_lines=None, split_on_newlines=True):
    """Reads files matching a wildcard pattern, yielding the contents.

    Args:
        filepattern: A wildcard pattern matching one or more files.
        max_lines: If set, stop reading after reading this many lines.
        split_on_newlines: A boolean. If true, then split files by lines and strip
                leading and trailing whitespace from each line. Otherwise, treat each
                file as a single string.

    Yields:
        The contents of the files as lines, if split_on_newlines is True, or
        the entire contents of each file if False.
    """
    filenames = sorted(tf.gfile.Glob(filepattern))
    lines_read = 0
    for filename in filenames:
        with tf.gfile.Open(filename) as f:
            if split_on_newlines:
                for line in f:
                    yield line.strip()
                    lines_read += 1
                    if max_lines and lines_read >= max_lines:
                        return

            else:
                if max_lines:
                    doc = []
                    for line in f:
                        doc.append(line)
                        lines_read += 1
                        if max_lines and lines_read >= max_lines:
                            yield "".join(doc)
                            return
                    yield "".join(doc)

                else:
                    yield f.read()


def corpus_token_counts(
        text_filepattern, corpus_max_lines, split_on_newlines=True):
    """Read the corpus and compute a dictionary of token counts.

    Args:
        text_filepattern: A pattern matching one or more files.
        corpus_max_lines: An integer; maximum total lines to read.
        split_on_newlines: A boolean. If true, then split files by lines and strip
                leading and trailing whitespace from each line. Otherwise, treat each
                file as a single string.

    Returns:
        a dictionary mapping token to count.
    """
    counts = collections.Counter()
    for doc in _read_filepattern(
            text_filepattern,
            max_lines=corpus_max_lines,
            split_on_newlines=split_on_newlines):
        counts.update(encode(_native_to_unicode(doc)))

    return counts


def vocab_token_counts(text_filepattern, max_lines):
    """Read a vocab file and return a dictionary of token counts.

    Reads a two-column CSV file of tokens and their frequency in a dataset. The
    tokens are presumed to be generated by encode() or the equivalent.

    Args:
        text_filepattern: A pattern matching one or more files.
        max_lines: An integer; maximum total lines to read.

    Returns:
        a dictionary mapping token to count.
    """
    ret = {}
    for i, line in enumerate(
            _read_filepattern(text_filepattern, max_lines=max_lines)):
        if "," not in line:
            tf.logging.warning("Malformed vocab line #%d '%s'", i, line)
            continue

        token, count = line.rsplit(",", 1)
        ret[_native_to_unicode(token)] = int(count)

    return ret



"""
Speech-to-text pipeline
"""

from at16k.core.media import Media
from at16k.core.model import Model
from at16k.core.segment import Segment
from at16k.blocks.transcriber import Transcriber


class SpeechToText():
    """
    Speech-to-text
    """

    def __init__(self, model_name):
        self._model = self._load_model(model_name)

    @staticmethod
    def _load_model(model_name):
        model = Model(model_name)
        return model

    def __call__(self, file_path):
        model = self._model
        media = Media(file_path)
        segment = Segment(waveform=media.waveform,
                          boundaries=(0.0, media.duration),
                          sample_rate=media.sample_rate,
                          channel=0)
        transcriber = Transcriber(model)
        text, score = transcriber(segment)
        result = {
            'file': file_path,
            'sample_rate': media.sample_rate,
            'begin': segment.boundaries[0],
            'end': segment.boundaries[1],
            'channel': segment.channel,
            'text': text,
            'score': score
        }
        return result


# coding=utf-8
# Copyright 2019 The Tensor2Tensor Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Encoders for text data.

* TextEncoder: base class
* ByteTextEncoder: for ascii text
* TokenTextEncoder: with user-supplied vocabulary file
* SubwordTextEncoder: invertible
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
from itertools import chain
import re
import time
import six
from six.moves import range    # pylint: disable=redefined-builtin
from at16k.utils import tokenizer

import tensorflow as tf

# Reserved tokens for things like padding and EOS symbols.
PAD = "<pad>"
EOS = "<EOS>"
RESERVED_TOKENS = [PAD, EOS]
NUM_RESERVED_TOKENS = len(RESERVED_TOKENS)
PAD_ID = RESERVED_TOKENS.index(PAD)    # Normally 0
EOS_ID = RESERVED_TOKENS.index(EOS)    # Normally 1

if six.PY2:
    RESERVED_TOKENS_BYTES = RESERVED_TOKENS
else:
    RESERVED_TOKENS_BYTES = [bytes(PAD, "ascii"), bytes(EOS, "ascii")]

# Regular expression for unescaping token strings.
# '\u' is converted to '_'
# '\\' is converted to '\'
# '\213;' is converted to unichr(213)
_UNESCAPE_REGEX = re.compile(r"\\u|\\\\|\\([0-9]+);")
_ESCAPE_CHARS = set(u"\\_u;0123456789")


# Unicode utility functions that work with Python 2 and 3
def native_to_unicode(s):
    if is_unicode(s):
        return s
    try:
        return to_unicode(s)
    except UnicodeDecodeError:
        res = to_unicode(s, ignore_errors=True)
        tf.logging.info("Ignoring Unicode error, outputting: %s" % res)
        return res


def unicode_to_native(s):
    if six.PY2:
        return s.encode("utf-8") if is_unicode(s) else s
    else:
        return s


def is_unicode(s):
    return isinstance(s, six.text_type)


def to_unicode(s, ignore_errors=False):
    if is_unicode(s):
        return s
    error_mode = "ignore" if ignore_errors else "strict"
    return s.decode("utf-8", errors=error_mode)


def to_unicode_ignore_errors(s):
    return to_unicode(s, ignore_errors=True)


def to_unicode_utf8(s):
    return str(s, "utf-8") if six.PY2 else s.decode("utf-8")


def strip_ids(ids, ids_to_strip):
    """Strip ids_to_strip from the end ids."""
    ids = list(ids)
    while ids and ids[-1] in ids_to_strip:
        ids.pop()
    return ids


class TextEncoder(object):
    """Base class for converting from ints to/from human readable strings."""

    def __init__(self, num_reserved_ids=NUM_RESERVED_TOKENS):
        self._num_reserved_ids = num_reserved_ids

    @property
    def num_reserved_ids(self):
        return self._num_reserved_ids

    def encode(self, s):
        """Transform a human-readable string into a sequence of int ids.

        The ids should be in the range [num_reserved_ids, vocab_size). Ids [0,
        num_reserved_ids) are reserved.

        EOS is not appended.

        Args:
            s: human-readable string to be converted.

        Returns:
            ids: list of integers
        """
        return [int(w) + self._num_reserved_ids for w in s.split()]

    def decode(self, ids, strip_extraneous=False):
        """Transform a sequence of int ids into a human-readable string.

        EOS is not expected in ids.

        Args:
            ids: list of integers to be converted.
            strip_extraneous: bool, whether to strip off extraneous tokens
                (EOS and PAD).

        Returns:
            s: human-readable string.
        """
        if strip_extraneous:
            ids = strip_ids(ids, list(range(self._num_reserved_ids or 0)))
        return " ".join(self.decode_list(ids))

    def decode_list(self, ids):
        """Transform a sequence of int ids into a their string versions.

        This method supports transforming individual input/output ids to their
        string versions so that sequence to/from text conversions can be visualized
        in a human readable format.

        Args:
            ids: list of integers to be converted.

        Returns:
            strs: list of human-readable string.
        """
        decoded_ids = []
        for id_ in ids:
            if 0 <= id_ < self._num_reserved_ids:
                decoded_ids.append(RESERVED_TOKENS[int(id_)])
            else:
                decoded_ids.append(id_ - self._num_reserved_ids)
        return [str(d) for d in decoded_ids]

    @property
    def vocab_size(self):
        raise NotImplementedError()


def _escape_token(token, alphabet):
    """Escape away underscores and OOV characters and append '_'.

    This allows the token to be expressed as the concatenation of a list
    of subtokens from the vocabulary. The underscore acts as a sentinel
    which allows us to invertibly concatenate multiple such lists.

    Args:
        token: A unicode string to be escaped.
        alphabet: A set of all characters in the vocabulary's alphabet.

    Returns:
        escaped_token: An escaped unicode string.

    Raises:
        ValueError: If the provided token is not unicode.
    """
    if not isinstance(token, six.text_type):
        raise ValueError(
            "Expected string type for token, got %s" % type(token))

    token = token.replace(u"\\", u"\\\\").replace(u"_", u"\\u")
    ret = [c if c in alphabet and c !=
           u"\n" else r"\%d;" % ord(c) for c in token]
    return u"".join(ret) + "_"


def _unescape_token(escaped_token):
    """Inverse of _escape_token().

    Args:
        escaped_token: a unicode string

    Returns:
        token: a unicode string
    """

    def match(m):
        if m.group(1) is None:
            return u"_" if m.group(0) == u"\\u" else u"\\"

        try:
            return six.unichr(int(m.group(1)))
        except (ValueError, OverflowError) as _:
            return u"\u3013"    # Unicode for undefined character.

    trimmed = escaped_token[:-
                            1] if escaped_token.endswith("_") else escaped_token
    return _UNESCAPE_REGEX.sub(match, trimmed)


class SubwordTextEncoder(TextEncoder):
    """Class for invertibly encoding text using a limited vocabulary.

    Invertibly encodes a native string as a sequence of subtokens from a limited
    vocabulary.

    A SubwordTextEncoder is built from a corpus (so it is tailored to the text in
    the corpus), and stored to a file. See text_encoder_build_subword.py.

    It can then be loaded and used to encode/decode any text.

    Encoding has four phases:

    1. Tokenize into a list of tokens.    Each token is a unicode string of either
         all alphanumeric characters or all non-alphanumeric characters.    We drop
         tokens consisting of a single space that are between two alphanumeric
         tokens.

    2. Escape each token.    This escapes away special and out-of-vocabulary
         characters, and makes sure that each token ends with an underscore, and
         has no other underscores.

    3. Represent each escaped token as a the concatenation of a list of subtokens
         from the limited vocabulary.    Subtoken selection is done greedily from
         beginning to end.    That is, we construct the list in order, always picking
         the longest subtoken in our vocabulary that matches a prefix of the
         remaining portion of the encoded token.

    4. Concatenate these lists.    This concatenation is invertible due to the
         fact that the trailing underscores indicate when one list is finished.

    """

    def __init__(self, filename=None):
        """Initialize and read from a file, if provided.

        Args:
            filename: filename from which to read vocab. If None, do not load a
                vocab
        """
        self._alphabet = set()
        self.filename = filename
        if filename is not None:
            self._load_from_file(filename)
        super(SubwordTextEncoder, self).__init__()

    def encode(self, s):
        """Converts a native string to a list of subtoken ids.

        Args:
            s: a native string.
        Returns:
            a list of integers in the range [0, vocab_size)
        """
        return self._tokens_to_subtoken_ids(
            tokenizer.encode(native_to_unicode(s)))

    def encode_without_tokenizing(self, token_text):
        """Converts string to list of subtoken ids without calling tokenizer.

        This treats `token_text` as a single token and directly converts it
        to subtoken ids. This may be useful when the default tokenizer doesn't
        do what we want (e.g., when encoding text with tokens composed of lots of
        nonalphanumeric characters). It is then up to the caller to make sure that
        raw text is consistently converted into tokens. Only use this if you are
        sure that `encode` doesn't suit your needs.

        Args:
            token_text: A native string representation of a single token.
        Returns:
            A list of subword token ids; i.e., integers in the range [0, vocab_size).
        """
        return self._tokens_to_subtoken_ids([native_to_unicode(token_text)])

    def decode(self, ids, strip_extraneous=False):
        """Converts a sequence of subtoken ids to a native string.

        Args:
            ids: a list of integers in the range [0, vocab_size)
            strip_extraneous: bool, whether to strip off extraneous tokens
                (EOS and PAD).

        Returns:
            a native string
        """
        if strip_extraneous:
            ids = strip_ids(ids, list(range(self._num_reserved_ids or 0)))
        return unicode_to_native(
            tokenizer.decode(self._subtoken_ids_to_tokens(ids)))

    def decode_list(self, ids):
        return [self._subtoken_id_to_subtoken_string(s) for s in ids]

    @property
    def vocab_size(self):
        """The subtoken vocabulary size."""
        return len(self._all_subtoken_strings)

    def _tokens_to_subtoken_ids(self, tokens):
        """Converts a list of tokens to a list of subtoken ids.

        Args:
            tokens: a list of strings.
        Returns:
            a list of integers in the range [0, vocab_size)
        """
        ret = []
        for token in tokens:
            ret.extend(self._token_to_subtoken_ids(token))
        return ret

    def _token_to_subtoken_ids(self, token):
        """Converts token to a list of subtoken ids.

        Args:
            token: a string.
        Returns:
            a list of integers in the range [0, vocab_size)
        """
        cache_location = hash(token) % self._cache_size
        cache_key, cache_value = self._cache[cache_location]
        if cache_key == token:
            return cache_value
        ret = self._escaped_token_to_subtoken_ids(
            _escape_token(token, self._alphabet))
        self._cache[cache_location] = (token, ret)
        return ret

    def _subtoken_ids_to_tokens(self, subtokens):
        """Converts a list of subtoken ids to a list of tokens.

        Args:
            subtokens: a list of integers in the range [0, vocab_size)
        Returns:
            a list of strings.
        """
        concatenated = "".join(
            [self._subtoken_id_to_subtoken_string(s) for s in subtokens])
        split = concatenated.split("_")
        ret = []
        for t in split:
            if t:
                unescaped = _unescape_token(t + "_")
                if unescaped:
                    ret.append(unescaped)
        return ret

    def _subtoken_id_to_subtoken_string(self, subtoken):
        """Converts a subtoken integer ID to a subtoken string."""
        if 0 <= subtoken < self.vocab_size:
            return self._all_subtoken_strings[subtoken]
        return u""

    def _escaped_token_to_subtoken_strings(self, escaped_token):
        """Converts an escaped token string to a list of subtoken strings.

        Args:
            escaped_token: An escaped token as a unicode string.
        Returns:
            A list of subtokens as unicode strings.
        """
        # NOTE: This algorithm is greedy; it won't necessarily produce the "best"
        # list of subtokens.
        ret = []
        start = 0
        token_len = len(escaped_token)
        while start < token_len:
            for end in range(
                    min(token_len, start + self._max_subtoken_len), start, -1):
                subtoken = escaped_token[start:end]
                if subtoken in self._subtoken_string_to_id:
                    ret.append(subtoken)
                    start = end
                    break

            else:    # Did not break
                # If there is no possible encoding of the escaped token then one of the
                # characters in the token is not in the alphabet. This should be
                # impossible and would be indicative of a bug.
                assert False, "Token substring not found in subtoken vocabulary."

        return ret

    def _escaped_token_to_subtoken_ids(self, escaped_token):
        """Converts an escaped token string to a list of subtoken IDs.

        Args:
            escaped_token: An escaped token as a unicode string.
        Returns:
            A list of subtoken IDs as integers.
        """
        return [
            self._subtoken_string_to_id[subtoken]
            for subtoken in self._escaped_token_to_subtoken_strings(escaped_token)
        ]

    @classmethod
    def build_from_generator(cls,
                             generator,
                             target_size,
                             max_subtoken_length=None,
                             reserved_tokens=None):
        """Builds a SubwordTextEncoder from the generated text.

        Args:
            generator: yields text.
            target_size: int, approximate vocabulary size to create.
            max_subtoken_length: Maximum length of a subtoken. If this is not set,
                then the runtime and memory use of creating the vocab is quadratic in
                the length of the longest token. If this is set, then it is instead
                O(max_subtoken_length * length of longest token).
            reserved_tokens: List of reserved tokens. The global variable
                `RESERVED_TOKENS` must be a prefix of `reserved_tokens`. If this
                argument is `None`, it will use `RESERVED_TOKENS`.

        Returns:
            SubwordTextEncoder with `vocab_size` approximately `target_size`.
        """
        token_counts = collections.defaultdict(int)
        for item in generator:
            for tok in tokenizer.encode(native_to_unicode(item)):
                token_counts[tok] += 1
        encoder = cls.build_to_target_size(
            target_size, token_counts, 1, 1e3,
            max_subtoken_length=max_subtoken_length,
            reserved_tokens=reserved_tokens)
        return encoder

    @classmethod
    def build_to_target_size(cls,
                             target_size,
                             token_counts,
                             min_val,
                             max_val,
                             max_subtoken_length=None,
                             reserved_tokens=None,
                             num_iterations=4):
        """Builds a SubwordTextEncoder that has `vocab_size` near `target_size`.

        Uses simple recursive binary search to find a minimum token count that most
        closely matches the `target_size`.

        Args:
            target_size: Desired vocab_size to approximate.
            token_counts: A dictionary of token counts, mapping string to int.
            min_val: An integer; lower bound for the minimum token count.
            max_val: An integer; upper bound for the minimum token count.
            max_subtoken_length: Maximum length of a subtoken. If this is not set,
                then the runtime and memory use of creating the vocab is quadratic in
                the length of the longest token. If this is set, then it is instead
                O(max_subtoken_length * length of longest token).
            reserved_tokens: List of reserved tokens. The global variable
                `RESERVED_TOKENS` must be a prefix of `reserved_tokens`. If this
                argument is `None`, it will use `RESERVED_TOKENS`.
            num_iterations: An integer; how many iterations of refinement.

        Returns:
            A SubwordTextEncoder instance.

        Raises:
            ValueError: If `min_val` is greater than `max_val`.
        """
        if min_val > max_val:
            raise ValueError("Lower bound for the minimum token count "
                             "is greater than the upper bound.")
        if target_size < 1:
            raise ValueError("Target size must be positive.")

        if reserved_tokens is None:
            reserved_tokens = RESERVED_TOKENS

        def bisect(min_val, max_val):
            """Bisection to find the right size."""
            present_count = (max_val + min_val) // 2
            tf.logging.info("Trying min_count %d" % present_count)
            subtokenizer = cls()
            subtokenizer.build_from_token_counts(
                token_counts, present_count, num_iterations,
                max_subtoken_length=max_subtoken_length,
                reserved_tokens=reserved_tokens)

            # Being within 1% of the target size is ok.
            is_ok = abs(subtokenizer.vocab_size -
                        target_size) * 100 < target_size
            # If min_val == max_val, we can't do any better than this.
            if is_ok or min_val >= max_val or present_count < 2:
                return subtokenizer

            if subtokenizer.vocab_size > target_size:
                other_subtokenizer = bisect(present_count + 1, max_val)
            else:
                other_subtokenizer = bisect(min_val, present_count - 1)

            if other_subtokenizer is None:
                return subtokenizer

            if (abs(other_subtokenizer.vocab_size - target_size) <
                    abs(subtokenizer.vocab_size - target_size)):
                return other_subtokenizer
            return subtokenizer

        return bisect(min_val, max_val)

    def build_from_token_counts(self,
                                token_counts,
                                min_count,
                                num_iterations=4,
                                reserved_tokens=None,
                                max_subtoken_length=None):
        """Train a SubwordTextEncoder based on a dictionary of word counts.

        Args:
            token_counts: a dictionary of Unicode strings to int.
            min_count: an integer - discard subtokens with lower counts.
            num_iterations: an integer.    how many iterations of refinement.
            reserved_tokens: List of reserved tokens. The global variable
                `RESERVED_TOKENS` must be a prefix of `reserved_tokens`. If this
                argument is `None`, it will use `RESERVED_TOKENS`.
            max_subtoken_length: Maximum length of a subtoken. If this is not set,
                then the runtime and memory use of creating the vocab is quadratic in
                the length of the longest token. If this is set, then it is instead
                O(max_subtoken_length * length of longest token).

        Raises:
            ValueError: if reserved is not 0 or len(RESERVED_TOKENS). In this case, it
                is not clear what the space is being reserved for, or when it will be
                filled in.
        """
        if reserved_tokens is None:
            reserved_tokens = RESERVED_TOKENS
        else:
            # There is not complete freedom in replacing RESERVED_TOKENS.
            for default, proposed in zip(RESERVED_TOKENS, reserved_tokens):
                if default != proposed:
                    raise ValueError("RESERVED_TOKENS must be a prefix of "
                                     "reserved_tokens.")

        # Initialize the alphabet. Note, this must include reserved tokens or it can
        # result in encoding failures.
        alphabet_tokens = chain(six.iterkeys(token_counts),
                                [native_to_unicode(t) for t in reserved_tokens])

        self._init_alphabet_from_tokens(alphabet_tokens)

        # Bootstrap the initial list of subtokens with the characters from the
        # alphabet plus the escaping characters.
        self._init_subtokens_from_list(list(self._alphabet),
                                       reserved_tokens=reserved_tokens)

        # We build iteratively.    On each iteration, we segment all the words,
        # then count the resulting potential subtokens, keeping the ones
        # with high enough counts for our new vocabulary.
        if min_count < 1:
            min_count = 1
        for i in range(num_iterations):
            tf.logging.info("Iteration {0}".format(i))

            # Collect all substrings of the encoded token that break along current
            # subtoken boundaries.
            subtoken_counts = collections.defaultdict(int)
            for token, count in six.iteritems(token_counts):
                iter_start_time = time.time()
                escaped_token = _escape_token(token, self._alphabet)
                subtokens = self._escaped_token_to_subtoken_strings(
                    escaped_token)
                start = 0
                for subtoken in subtokens:
                    last_position = len(escaped_token) + 1
                    if max_subtoken_length is not None:
                        last_position = min(
                            last_position, start + max_subtoken_length)

                    for end in range(start + 1, last_position):
                        new_subtoken = escaped_token[start:end]
                        subtoken_counts[new_subtoken] += count
                    start += len(subtoken)
                iter_time_secs = time.time() - iter_start_time
                if iter_time_secs > 0.1:
                    tf.logging.info(u"Processing token [{0}] took {1} seconds, consider "
                                    "setting Text2TextProblem.max_subtoken_length to a "
                                    "smaller value.".format(token, iter_time_secs))

            # Array of sets of candidate subtoken strings, by length.
            len_to_subtoken_strings = []
            for subtoken_string, count in six.iteritems(subtoken_counts):
                lsub = len(subtoken_string)
                if count >= min_count:
                    while len(len_to_subtoken_strings) <= lsub:
                        len_to_subtoken_strings.append(set())
                    len_to_subtoken_strings[lsub].add(subtoken_string)

            # Consider the candidates longest to shortest, so that if we accept
            # a longer subtoken string, we can decrement the counts of its prefixes.
            new_subtoken_strings = []
            for lsub in range(len(len_to_subtoken_strings) - 1, 0, -1):
                subtoken_strings = len_to_subtoken_strings[lsub]
                for subtoken_string in subtoken_strings:
                    count = subtoken_counts[subtoken_string]
                    if count >= min_count:
                        # Exclude alphabet tokens here, as they must be included later,
                        # explicitly, regardless of count.
                        if subtoken_string not in self._alphabet:
                            new_subtoken_strings.append(
                                (count, subtoken_string))
                        for l in range(1, lsub):
                            subtoken_counts[subtoken_string[:l]] -= count

            # Include the alphabet explicitly to guarantee all strings are encodable.
            new_subtoken_strings.extend((subtoken_counts.get(a, 0), a)
                                        for a in self._alphabet)
            new_subtoken_strings.sort(reverse=True)

            # Reinitialize to the candidate vocabulary.
            new_subtoken_strings = [subtoken for _,
                                    subtoken in new_subtoken_strings]
            if reserved_tokens:
                escaped_reserved_tokens = [
                    _escape_token(native_to_unicode(t), self._alphabet)
                    for t in reserved_tokens
                ]
                new_subtoken_strings = escaped_reserved_tokens + new_subtoken_strings

            self._init_subtokens_from_list(new_subtoken_strings)
            tf.logging.info("vocab_size = %d" % self.vocab_size)

    @property
    def all_subtoken_strings(self):
        return tuple(self._all_subtoken_strings)

    def dump(self):
        """Debugging dump of the current subtoken vocabulary."""
        subtoken_strings = [(i, s)
                            for s, i in six.iteritems(self._subtoken_string_to_id)]
        print(u", ".join(u"{0} : '{1}'".format(i, s)
                         for i, s in sorted(subtoken_strings)))

    def _init_subtokens_from_list(self, subtoken_strings, reserved_tokens=None):
        """Initialize token information from a list of subtoken strings.

        Args:
            subtoken_strings: a list of subtokens
            reserved_tokens: List of reserved tokens. We must have `reserved_tokens`
                as None or the empty list, or else the global variable `RESERVED_TOKENS`
                must be a prefix of `reserved_tokens`.

        Raises:
            ValueError: if reserved is not 0 or len(RESERVED_TOKENS). In this case, it
                is not clear what the space is being reserved for, or when it will be
                filled in.
        """
        if reserved_tokens is None:
            reserved_tokens = []

        if reserved_tokens:
            self._all_subtoken_strings = reserved_tokens + subtoken_strings
        else:
            self._all_subtoken_strings = subtoken_strings

        # we remember the maximum length of any subtoken to avoid having to
        # check arbitrarily long strings.
        self._max_subtoken_len = max([len(s) for s in subtoken_strings])
        self._subtoken_string_to_id = {
            s: i + len(reserved_tokens)
            for i, s in enumerate(subtoken_strings) if s
        }
        # Initialize the cache to empty.
        self._cache_size = 2 ** 20
        self._cache = [(None, None)] * self._cache_size

    def _init_alphabet_from_tokens(self, tokens):
        """Initialize alphabet from an iterable of token or subtoken strings."""
        # Include all characters from all tokens in the alphabet to guarantee that
        # any token can be encoded. Additionally, include all escaping characters.
        self._alphabet = {c for token in tokens for c in token}
        self._alphabet |= _ESCAPE_CHARS

    def _load_from_file_object(self, f):
        """Load from a file object.

        Args:
            f: File object to load vocabulary from
        """
        subtoken_strings = []
        for line in f:
            s = line.rstrip()
            # Some vocab files wrap words in single quotes, but others don't
            if ((s.startswith("'") and s.endswith("'")) or
                    (s.startswith("\"") and s.endswith("\""))):
                s = s[1:-1]
            subtoken_strings.append(native_to_unicode(s))
        self._init_subtokens_from_list(subtoken_strings)
        self._init_alphabet_from_tokens(subtoken_strings)

    def _load_from_file(self, filename):
        """Load from a vocab file."""
        if not tf.io.gfile.exists(filename):
            raise ValueError("File %s not found" % filename)
        with tf.io.gfile.GFile(filename) as f:
            self._load_from_file_object(f)

    def store_to_file(self, filename, add_single_quotes=True):
        with tf.gfile.Open(filename, "w") as f:
            for subtoken_string in self._all_subtoken_strings:
                if add_single_quotes:
                    f.write("'" + unicode_to_native(subtoken_string) + "'\n")
                else:
                    f.write(unicode_to_native(subtoken_string) + "\n")



"""
Speech to text converter (command-line)
"""
import argparse
from at16k.api import SpeechToText

PARSER = argparse.ArgumentParser('at16k Speech-to-Text')
PARSER.add_argument('-i', '--input', type=str,
                    required=True, help='Input file (wav)')
PARSER.add_argument('-m', '--model', type=str,
                    required=True, choices=['en_8k', 'en_16k'])
FLAGS = PARSER.parse_args()

def main():
    """
    Main
    """
    stt = SpeechToText(FLAGS.model)
    result = stt(FLAGS.input)
    print('-'*100)
    print('Output:', result['text'])
    print('-'*100)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['at16k', 'at16k.api']

package_data = \
{'': ['*'], 'at16k': ['bin/*', 'blocks/*', 'core/*', 'utils/*']}

install_requires = \
['progressbar>=2.5,<3.0', 'scipy>=1.3.3,<2.0.0', 'tensorflow==1.14']

entry_points = \
{'console_scripts': ['at16k-convert = at16k.bin.speech_to_text:main']}

setup_kwargs = {
    'name': 'at16k',
    'version': '0.1.1',
    'description': 'at16k is a Python library to perform automatic speech recognition or speech to text conversion.',
    'long_description': "# at16k\nPronounced as ***at sixteen k***\n\n# What is at16k?\nat16k is a Python library to perform automatic speech recognition or speech to text conversion. The goal of this project is to provide the community with a production quality speech-to-text library.\n\n# Installation\nIt is recommended that you install at16k in a virtual environment.\n\n## Prerequisites\n- Python = 3.6 (not tested on other versions)\n- Tensorflow = 1.14\n- Scipy (for reading wav files)\n\n## Install via pip\n```\n$ pip install at16k\n```\n\n## Install from source\nRequires: [poetry](https://github.com/sdispater/poetry)\n```\n$ git clone https://github.com/at16k/at16k.git\n$ poetry env use python3.6\n$ poetry install\n```\n\n# Download models\nCurrently, two models are available for speech to text conversion.\n- en_8k (Trained on english audio recorded at 8 KHz)\n- en_16k (Trained on english audio recorded at 16 KHz)\n\nTo download all the models:\n```\n$ python -m at16k.download all\n```\nAlternatively, you can download only the model you need. For example:\n```\n$ python -m at16k.download en_8k\n$ python -m at16k.download en_16k\n```\n\n# Preprocessing audio files\nat16k accepts wav files with the following spces:\n- Channels: 1\n- Bits per sample: 16\n- Sample rate: 8000 (en_8k) or 16000 (en_16k)\n\nUse ffmpeg to convert your audio/video files to an acceptable format. For example,\n```\n# For 8 KHz\n$ ffmpeg -i <input_file> -ar 8000 -ac 1 -ab 16 <output_file>\n\n# For 16 KHz\n$ ffmpeg -i <input_file> -ar 16000 -ac 1 -ab 16 <output_file>\n```\n\n# Usage\n\n## Command line\nThere are two ways to invoke at16k speech-to-text via the command line.\n```\nat16k-convert -i <input_wav_file> -m <model_name>\n```\nAlternatively,\n```\npython -m at16k.bin.speech_to_text -i <input_wav_file> -m <model_name>\n```\n## Library API\n```\nfrom at16k.api import SpeechToText\n\n# One-time initialization\nSTT = SpeechToText('en_16k') # or en_8k\n\n# Run STT on an audio file, returns a dict\nprint(STT('./samples/test_16k.wav'))\n```\nCheck [example.py](example.py) for details on how to use the API.\n\n# Limitations\n\nThe max duration of your audio file should be less than **30 seconds** when using **en_8k**, and less than **15 seconds** when using **en_16k**. An error will not be thrown ff the duration exceeds the limits, however, your transcript may contain errors and missing text.\n",
    'author': 'Mohit Shah',
    'author_email': 'mohit@at16k.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/at16k/at16k.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
Transcriber
"""

import tensorflow as tf
from at16k.core.segment import Segment
from at16k.core.model import Model

class Transcriber():
    """
    Transcribe an audio segment

    Returns:
        [type] -- [description]
    """

    def __init__(self, model: Model):
        self._model = model

    @staticmethod
    def _make_example(waveform):
        features = {
            "waveforms": tf.train.Feature(float_list=tf.train.FloatList(value=waveform))
        }
        example = tf.train.Example(
            features=tf.train.Features(feature=features))
        return example.SerializeToString()

    @staticmethod
    def _decode_example(example, model):
        pred_fn = model.pred_fn
        output_fn = model.output_fn
        predictions = pred_fn({'input': [example]})
        output = list(predictions['outputs'])[0]
        score = list(predictions['scores'])[0]
        text = output_fn(output, strip_extraneous=True)
        return text, float(score)

    def __call__(self, segment: Segment):
        model = self._model
        waveform = segment.waveform
        example = self._make_example(waveform)
        text, score = self._decode_example(example, model)
        return text, score
